/*
 *   stunnel       Universal SSL tunnel
 *   Copyright (c) 1998-2000 Michal Trojnara <Michal.Trojnara@centertel.pl>
 *                 All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*	klpstunnel  stunnel-3.8, modified for use as kerberized local web proxy
**	20000310	VRS 	Started.  Changes to transfer() function.
*/
#define	PROXY_VERSION "stunnel-3.8/kssl-0.2"

const char	*connect_response =
		"HTTP/1.0 200 Connection established\n"
		"Proxy-agent: "  PROXY_VERSION  "\n\n";

int 		have_cookie = 0;


/* For US citizens having problems with patents, undefined by default */
/* Uncomment the next line to disable RSA support */
/* #define NO_RSA */

/* Experimental DH support is disabled by default */
/* Comment out the next line if you need it */
#define NO_DH

/* Non-blocking sockets are disabled by default */
/* It works on most systems so feel free to uncomment the next line */
/* #define USE_NBIO */

/* Length of temporary RSA key */
#ifndef NO_RSA
#define KEYLENGTH      512
#endif /* NO_RSA */

/* Undefine if you have problems with make_sockets() */
#define INET_SOCKET_PAIR

/* Max number of children */
#define MAX_CLIENTS    100

#include "common.h"

#include <stdio.h>

#ifdef USE_WIN32

#define Win32_Winsock
#include <windows.h>
#include <io.h>
#include <stdlib.h>
#include <process.h>
#include <malloc.h>
#include <string.h>
#include <sys/stat.h>
#define ECONNRESET WSAECONNRESET
#define ENOTSOCK WSAENOTSOCK

#else /* defined USE_WIN32 */

/* Must be included before sys/stat.h for Ultrix */
#include <sys/types.h>   /* u_short, u_long */

/* General headers */
#include <errno.h>       /* errno */
#include <sys/stat.h>    /* stat */
#include <signal.h>      /* signal */
#include <sys/wait.h>    /* wait */
#include <string.h>      /* strerror */
#include <ctype.h>
#include <stdlib.h>
#include <netdb.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>      /* getopt */
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>      /* getpid, fork, execv, exit */
#endif

/* Networking headers */
#include <netinet/in.h>  /* struct sockaddr_in */
#include <sys/socket.h>  /* getpeername */
#include <arpa/inet.h>   /* inet_ntoa */
#include <sys/time.h>    /* select */
#include <sys/ioctl.h>   /* ioctl */
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>  /* for aix */
#endif

    /* TCP wrapper */
#ifdef USE_LIBWRAP
#include <tcpd.h>
int allow_severity=LOG_NOTICE;
int deny_severity=LOG_WARNING;
#endif

#endif /* defined USE_WIN32 */

extern server_options options;

int auth_user(struct sockaddr_in *);

    /* SSL functions */
void client(int);
static int transfer(Channel *local, Channel *remote);
#ifndef NO_RSA
static RSA *tmp_rsa_cb(SSL *, int, int);
#endif /* NO_RSA */
static int verify_callback (int, X509_STORE_CTX *);
static void info_callback(SSL *, int, int);
static void print_stats(SSL_CTX *ctx);
static void print_cipher(SSL *);
static void sslerror(char *);

void ssl3_send_alert(SSL *s,int level, int desc);

/* Correct callback definitions overriding ssl.h */
#ifndef NO_RSA
#ifdef SSL_CTX_set_tmp_rsa_callback
    #undef SSL_CTX_set_tmp_rsa_callback
#endif
#define SSL_CTX_set_tmp_rsa_callback(ctx,cb) \
    SSL_CTX_ctrl(ctx,SSL_CTRL_SET_TMP_RSA_CB,0,(char *)cb)
#endif /* NO_RSA */

#ifndef NO_RSA
RSA *rsa_tmp;           /* temporary RSA key */
#endif /* NO_RSA */

/*	Initialize & return SSL context.
**	May need args if same option.options don't work for both 
**	remote (normal) and local legs.  E.g. option.cipher_list?
*/
SSL_CTX	*context_init(char *host)
{
	SSL_CTX *ctx;           /* (no longer) global SSL context */
#ifndef NO_DH
    static DH *dh=NULL;
    BIO *bio=NULL;
#endif /* NO_DH */

    yaapp_RAND_load_file(NULL, NULL, 0);
    SSLeay_add_ssl_algorithms();
    SSL_load_error_strings();
    if ((options.option & OPT_KHTTPS)  &&  host  &&  strlen(host))  {
		ctx=SSL_CTX_new(SSLv23_client_method());
		}
	else if ((options.option & OPT_CLIENT)  &&  host  &&  strlen(host))  {
        ctx=SSL_CTX_new(SSLv3_client_method());
		}
	else  { /* Server mode */
        ctx=SSL_CTX_new(SSLv23_server_method());
#ifndef NO_RSA
        log(LOG_DEBUG, "Generating %d bit temporary RSA key...", KEYLENGTH);
#if SSLEAY_VERSION_NUMBER <= 0x0800
        rsa_tmp=RSA_generate_key(KEYLENGTH, RSA_F4, NULL);
#else
        rsa_tmp=RSA_generate_key(KEYLENGTH, RSA_F4, NULL, NULL);
#endif
        if(!rsa_tmp) {
            sslerror("tmp_rsa_cb");
            exit(1);
			}
        log(LOG_DEBUG, "Temporary RSA key generated");
        SSL_CTX_set_tmp_rsa_callback(ctx, tmp_rsa_cb);
#endif /* NO_RSA */
#ifndef NO_DH
        if(!(bio=BIO_new_file(options.certfile, "r"))) {
            log(LOG_ERR, "DH: Could not read %s: %s", options.certfile,
                strerror(errno));
            goto dh_failed;
        }
        if(!(dh=PEM_read_bio_DHparams(bio, NULL, NULL
#if SSLEAY_VERSION_NUMBER >= 0x00904000L
                , NULL
#endif
                ))) {
            log(LOG_ERR, "Could not load DH parameters from %s",
                options.certfile);
            goto dh_failed;
        }
        SSL_CTX_set_tmp_dh(ctx, dh);
        log(LOG_DEBUG, "Diffie-Hellman initialized with %d bit key",
            8*DH_size(dh));
        goto dh_done;
dh_failed:
        log(LOG_WARNING, "Diffie-Hellman initialization failed");
dh_done:
        if(bio)
            BIO_free(bio);
        if(dh)
            DH_free(dh);
#endif /* NO_DH */
    }
    SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_BOTH);
    SSL_CTX_set_timeout(ctx, options.session_timeout);
    if(options.option&OPT_CERT) {
        if(!SSL_CTX_use_certificate_file(ctx, options.certfile,
                SSL_FILETYPE_PEM)) {
            log(LOG_ERR, "Error reading certificate file: %s", options.certfile);
            sslerror("SSL_CTX_use_certificate_file");
            exit(1);
        }
        log(LOG_DEBUG, "Certificate: %s", options.certfile);
#ifdef NO_RSA
        if(!SSL_CTX_use_PrivateKey_file(ctx, options.certfile,
                SSL_FILETYPE_PEM)) {
            sslerror("SSL_CTX_use_PrivateKey_file");
            exit(1);
        }
#else /* NO_RSA */
        if(!SSL_CTX_use_RSAPrivateKey_file(ctx, options.certfile,
                SSL_FILETYPE_PEM)) {
            sslerror("SSL_CTX_use_RSAPrivateKey_file");
            exit(1);
        }
#endif /* NO_RSA */
    }
    if(options.verify_level!=SSL_VERIFY_NONE) {
        if ((!SSL_CTX_set_default_verify_paths(ctx))
                || (!SSL_CTX_load_verify_locations(ctx, CLIENT_CA,
                options.clientdir))){
            sslerror("X509_load_verify_locations");
            exit(1);
        }
        SSL_CTX_set_verify(ctx, options.verify_level, verify_callback);
        SSL_CTX_set_client_CA_list(ctx, SSL_load_client_CA_file(CLIENT_CA));
        if (options.verify_use_only_my)
            log(LOG_NOTICE, "Peer certificate location %s", options.clientdir);
    }
    SSL_CTX_set_info_callback(ctx, info_callback);
    if (options.loc_cipher_list  ||  options.rem_cipher_list) {
		char	*cipher_list = (host)?	options.rem_cipher_list:
										options.loc_cipher_list;
        if (cipher_list  &&  strlen(cipher_list)  &&
			!SSL_CTX_set_cipher_list(ctx, cipher_list))  {
            sslerror("SSL_CTX_set_cipher_list");
            exit(1);
        }
    }

	return ctx;
}

void client(int local)
{
    struct sockaddr_in addr;
    int addrlen;
    int remote;
    struct linger l;
    u_long ip;
#ifdef USE_LIBWRAP
    struct request_info request;
#endif
    Channel	remote_ch;						/*  Normally SSL	*/
	Channel	local_ch;						/*  Normally socket	*/

    log(LOG_DEBUG, "%s started", options.servname);
    l.l_onoff=1;
    l.l_linger=0;
    addrlen=sizeof(addr);
    if(getpeername(local, (struct sockaddr *)&addr, &addrlen)<0) {
        if(options.option&OPT_TRANSPARENT || errno!=ENOTSOCK) {
            sockerror("getpeerbyname");
            goto cleanup_local;
        }
        /* Ignore ENOTSOCK error so 'local' doesn't have to be a socket */
    } else {
#ifdef USE_LIBWRAP
        request_init(&request, RQ_DAEMON, options.servname, RQ_FILE, local, 0);
        fromhost(&request);
        if (!hosts_access(&request)) {
            log(LOG_WARNING, "Connection from %s:%d REFUSED by libwrap",
                inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
            log(LOG_DEBUG, "See hosts_access(5) for details");
            goto cleanup_local;
        }
#endif
        if(auth_user(&addr)<0) {
            log(LOG_WARNING, "Connection from %s:%d REFUSED by IDENT",
                inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
            goto cleanup_local;
        }
        log(LOG_NOTICE, "%s connected from %s:%d", options.servname,
            inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
    }

    /* create connection to host/service */
    ip=options.option&OPT_TRANSPARENT ? addr.sin_addr.s_addr : 0;
    if(options.option&OPT_REMOTE) { /* remote host */
        if((remote=connect_remote(ip))<0)
            goto cleanup_local; /* Failed to connect remote server */
        log(LOG_DEBUG, "Remote host connected");
    } else { /* local service */
        if((remote=connect_local(ip))<0)
            goto cleanup_local; /* Failed to spawn local service */
        log(LOG_DEBUG, "Local service connected");
    }

    /* negotiate protocol */
    if(negotiate(options.protocol, options.option&OPT_CLIENT,
            local, remote) <0) {
        log(LOG_ERR, "Protocol negotiations failed");
        goto cleanup_remote;
    }

	if (!init_ch(&remote_ch)  ||  !init_ch(&local_ch))  {
		log(LOG_ERR, "init_ch fails");
        goto cleanup_local;
		}
	else  {
		/*	200000316 VRS - issues w/ session spanning multiple transfer calls.
		**	Orig stunnel reset here, but always to same values (fixed channels)
		**	Need to leave wr_remote_sock fd alone across transfers...
		**	How do I know when to reset it; i.e. when session ends?
		**	20000317 VRS - moved up from transfer() to client()
		*/
		local_ch.fd  = (options.option&OPT_CLIENT)? local: remote;
		remote_ch.fd = (options.option&OPT_CLIENT)? remote: local;
		}

	if (transfer(&local_ch, &remote_ch) < 0)  goto cleanup_ssl;

    /* No error - normal shutdown */
	shutdown_ssl_ch(&local_ch, &remote_ch);
    closesocket(remote);
    closesocket(local);
    goto done;
cleanup_ssl: /* close SSL and reset sockets */
	shutdown_ssl_ch(&local_ch, &remote_ch);
cleanup_remote: /* reset remote and local socket */
    if ((options.option & OPT_REMOTE) &&
        setsockopt(remote, SOL_SOCKET, SO_LINGER, (char *)&l, sizeof(l)) < 0 &&
        errno != ENOTSOCK)
        sockerror("linger (remote)");
    closesocket(remote);
cleanup_local: /* reset local socket */
    if (!((options.option & OPT_CLIENT) && (options.option & OPT_PROGRAM)) &&
        setsockopt(local, SOL_SOCKET, SO_LINGER, (char *)&l, sizeof(l)) < 0 &&
        errno != ENOTSOCK)
        sockerror("linger (local)");
    closesocket(local);
done:
#ifndef USE_FORK
    log(LOG_DEBUG, "%s finished (%d left)", options.servname,
        --options.clients);
#endif
    ; /* ANSI C compiler needs it */
}


/*	Initialize/clear Channel struct.
**	Returns 0=problem; 1=OK.
*/
int 	init_ch(Channel *channel)
{
	if (channel == NULL)  return 0;

	channel->open	= CHAN_OPEN;
	channel->ctx	= NULL;
	channel->ssl	= NULL;
	channel->fd 	= 0;
	channel->bytes	= 0;
	channel->ptr	= 0;

	return 1;
}


void	show_sock(int s)
{
    struct sockaddr_in	addr, paddr;
	int 				addrlen, paddrlen;
	char				*self, *other;
	int 				sport, oport;

    if (getsockname(s, (struct sockaddr *) &addr, &addrlen) < 0)  {
		printf("show_sock(%d): getsockname failed\n", s);
		}
    if (getpeername(s, (struct sockaddr *) &paddr, &paddrlen) < 0)  {
		printf("show_sock(%d): getpeername failed\n", s);
		}
	
	self = inet_ntoa(addr.sin_addr);	other = inet_ntoa(paddr.sin_addr);
	sport= ntohs(addr.sin_port);		oport = ntohs(paddr.sin_port);

	if (!self)	self  = "undef";
	if (!other)	other = "undef";

	printf("Socket %d (%s:%d) is connected to: %s:%d\n", s,
					self,sport, other,oport);
}


#define	MAX_SESS	40;

/*	Create SSL connection on (local or remote) channel to given host.
**	Return 0:Success.
*/
int 	new_ssl_ch(Channel *chan, char *host)
{
	EVP_MD_CTX		hash_ctx;
	unsigned char	*sess_id;
	unsigned int 	len = getpid();

	if ((!chan->ctx  &&  !(chan->ctx = context_init(host)))  ||
		(!chan->ssl  &&  !(chan->ssl = SSL_new(chan->ctx))))  {
        sslerror("SSL_new");
        return -1;
		}
	/*
	SSL_clear(chan->ssl);
	*/

#if SSLEAY_VERSION_NUMBER >= 0x0922
	EVP_DigestInit(&hash_ctx, EVP_get_digestbyname("ssl3-sha1"));
	EVP_DigestUpdate(&hash_ctx, chan->ssl, sizeof(chan));
	EVP_DigestUpdate(&hash_ctx, chan, sizeof(chan));
	EVP_DigestUpdate(&hash_ctx, &len, sizeof(len));
	if (!(sess_id = (unsigned char *) malloc(EVP_MD_size(hash_ctx.digest))))  {
		sslerror("session_id_context alloc");
		return -1;
		}
	EVP_DigestFinal(&hash_ctx, sess_id, &len);
    SSL_set_session_id_context(chan->ssl, sess_id, len);
	free(sess_id);
#endif

#ifndef NO_KRB5
	if (chan->ssl->kssl_ctx)  {
		kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_SERVICE, KRB5SVC);
		kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_KEYTAB, KRB5KEYTAB);
		if (host)  kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_SERVER, host);
		}
#endif	/* NO_KRB5 */

	return 0;
}


/*	Close any SSL channels.
*/
void	shutdown_ssl_ch(Channel *local, Channel *remote)
{
	have_cookie = 0;

	if (local  &&  local->ssl)  {
		SSL_set_shutdown(local->ssl,  SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
		SSL_free(local->ssl);
		local->ssl = NULL;
		local->ctx = NULL;
		}
	if (remote  &&  remote->ssl)  {
		SSL_set_shutdown(remote->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
		SSL_free(remote->ssl);
		remote->ssl = NULL;
		remote->ctx = NULL;
		}
    ERR_remove_state(0);
}


/*	Set SSL fd and do SSL_connect.
**	Return 0:Success.
*/
int 	connect_ssl_ch(Channel *chan, char *hp)
{
	if (!chan->fd  &&  create_remote_sock(chan, hp))  {
		sslerror("ssl create remote sock");
		return 1;
		}
	SSL_set_fd(chan->ssl, chan->fd);
	SSL_set_connect_state(chan->ssl);
	if (SSL_connect(chan->ssl) <= 0)  {
		sslerror("SSL_connect");
		return 1;
		}
	/*
	print_cipher(chan->ssl);
	*/

	return 0;
}

/*	Set SSL fd and do SSL_accept.
**	Return 0:Success.
*/
int 	accept_ssl_ch(Channel *chan)
{
	SSL_set_fd(chan->ssl, chan->fd);
	SSL_set_accept_state(chan->ssl);
	if (SSL_accept(chan->ssl) <= 0)  {
		sslerror("SSL_accept");
		return 1;
		}
	/*
	print_cipher(chan->ssl);
	*/

	return 0;
}

/*	Do SSL read on given channel, return number of bytes read through *numptr.
**	Return 0:Success
*/
int 	rd_ssl_ch(Channel *chan, int *numptr)
{
	*numptr = SSL_read(chan->ssl, chan->buff + chan->ptr, BUFFSIZE - chan->ptr);

	switch (SSL_get_error(chan->ssl, *numptr))  {
		case SSL_ERROR_NONE:
			chan->ptr += *numptr;
			break;
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
			log(LOG_DEBUG, "SSL_read returned WANT_ - retry");
			break;
		case SSL_ERROR_SYSCALL:
			if (*numptr)  { 	/* not EOF */
				sockerror("SSL_read (socket)");
				return -1;
			}
		case SSL_ERROR_ZERO_RETURN:
			log(LOG_DEBUG, "SSL closed on read");
			chan->open = 0;
			break;
		case SSL_ERROR_SSL:
		default:
			sslerror("SSL_read");
			return -1;
		}

	return 0;
}

/*	Do SSL write on given channel, return #bytes written through *numptr.
**	Return 0:Success
*/
int 	wr_ssl_ch(Channel *from, Channel *to, int *numptr)
{
	*numptr = SSL_write(to->ssl, from->buff, from->ptr);

	switch (SSL_get_error(to->ssl, *numptr)) {
		case SSL_ERROR_NONE:
			memcpy(from->buff, from->buff + *numptr, from->ptr - *numptr);
			from->ptr -= *numptr;
			to->bytes += *numptr;
			break;
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
			log(LOG_DEBUG, "SSL_write returned WANT_ - retry");
			break;
		case SSL_ERROR_SYSCALL:
			if (*numptr)  { /* not EOF */
				sockerror("SSL_write (socket)");
				return -1;
				}
		case SSL_ERROR_ZERO_RETURN:
			log(LOG_DEBUG, "SSL closed on write");
			to->open = 0;
			break;
		case SSL_ERROR_SSL:
		default:
			sslerror("SSL_write");
			return -1;
		}

	return 0;
}


/*	Create and connect any local/remote sockets or ssl connections,
**	as appropriate for km_mode.  KM_NORMAL mode should match original stunnel.
**	The local socket connection should already be established.
**	Return 0:Success.
*/
int 	connect_ch(Channel *local, Channel *remote, int km_mode, char *hp)
{
	char	*cp, host[MAX_HP+1];
	int 	rc = 0;

	cp = (km_mode == KM_NORMAL)? options.servname: hp;
	if (strlen(cp) >= MAX_HP)  {
		sockerror("host name overflow");
		return -1;
		}
	strcpy(host, cp);
	cp = strrchr(host, (km_mode == KM_NORMAL)? '.': ':');
	if (cp)  *cp = '\0';

	switch (km_mode)  {

		case KM_NORMAL:
				/*	Add remote ssl connection to options.servname		*/
				new_ssl_ch(remote, host);

				if (options.option & OPT_CLIENT)  {
					if (!hp  ||  !strlen(hp))  {
						hp = (char *) options.localnames;
						if (!hp  ||  !strlen(hp))  hp = "localhost";
						snprintf(host, MAX_HP, "%s:%d", hp, options.localport);
						hp = host;
						}
					if (connect_ssl_ch(remote, hp))  return -1;
					}
				else  {
					if (accept_ssl_ch(remote))   return -1;
					}
				break;

		case KM_KSSL_MITM:
				/*	Add remote ssl connection to host,
				**	Add local ssl connection and use in place of local->fd
				*/
				rc = new_ssl_ch(remote, host);

				/* VRS 20000324 late: maybe blame existing remote->fd = 10 so...
				   if (remote->fd)  { closesocket(remote->fd); remote->fd=0; }
				*/
				  remote->fd = 0;
				if (connect_ssl_ch(remote, hp))  return -1;

				if (! fake_CONNECT(local))  return -1;

				new_ssl_ch(local, NULL);
				if (accept_ssl_ch(local))  return -1;

				break;

		case KM_LOCAL_HTTPS:
				/*	Add remote sock connection to host					*/
				if (create_remote_sock(remote, hp))  {
					sockerror("local https");
					return -1;
					}

				if (! fake_CONNECT(local))  return -1;
				break;

		case KM_REMOTE_HTTPS:
				/*	Add remote sock connection to host					*/
				if (create_remote_sock(remote, hp))  {
					sockerror("remote https");
					return -1;
					}
				break;

		default:
				return -1;
				break;
		}

	return 0;
}


static	fd_set	rd_set, wr_set;

/*	transfer data
*/
static int transfer(Channel *local, Channel *remote)
{
    int 	fdno;
	int 	retval = 1;
	int 	km_mode = KM_UNDEF;
#if defined FIONBIO && defined USE_NBIO
    unsigned long	l;
#endif

#if defined FIONBIO && defined USE_NBIO
	l=1; /* ON */
	if(ioctlsocket(local->fd, FIONBIO, &l)<0)
		sockerror("ioctlsocket (sock)"); /* non-critical */
	if(ioctlsocket(remote->fd, FIONBIO, &l)<0)
		sockerror("ioctlsocket (ssl)"); /* non-critical */
	log(LOG_DEBUG, "Sockets set to non-blocking mode");
#endif

    while( retval > 0  &&
		(local->open||local->ptr) && (remote->open||remote->ptr))  {

		fdno = ((remote->fd > local->fd)? remote->fd: local->fd) + 1;

        FD_ZERO(&rd_set);
        if(local->open && local->ptr<BUFFSIZE) /* can read from socket */
            if (local->fd)  FD_SET(local->fd, &rd_set);
        if(remote->open && remote->ptr<BUFFSIZE) /* can read from SSL */
            if (remote->fd)  FD_SET(remote->fd, &rd_set);

        FD_ZERO(&wr_set);
        if(local->open && remote->ptr) /* can write to socket */
            if (local->fd)  FD_SET(local->fd, &wr_set);
        if(remote->open && local->ptr) /* can write to SSL */
            if (remote->fd)  FD_SET(remote->fd, &wr_set);

        if(select(fdno, &rd_set, &wr_set, NULL, NULL)<0) {
            sockerror("select");
			retval = -1;
			break;
        }

		if ((retval = rd_local (local, remote, &km_mode)) <= 0)  break;
		if ((retval = rd_remote(local, remote, &km_mode)) <= 0)  break;
		if ((retval = got_cookie(local,remote,  km_mode)) <= 0)  break;
		if ((retval = wr_local (local, remote,  km_mode)) <= 0)  break;
		if ((retval = wr_remote(local, remote,  km_mode)) <= 0)  break;
    }

	have_cookie = 0;

#if defined FIONBIO && defined USE_NBIO
    l=0; /* OFF */
    if(ioctlsocket(local->fd, FIONBIO, &l)<0)
        sockerror("ioctlsocket (sock)"); /* non-critical */
    if(ioctlsocket(remote->fd, FIONBIO, &l)<0)
        sockerror("ioctlsocket (ssl)"); /* non-critical */
#endif

    log(LOG_NOTICE,
        "Connection %s: %d bytes sent to SSL, %d bytes sent to socket",
        retval<0 ? "reset" : "closed", remote->bytes, local->bytes);
    return retval;
}


/*	Search and destroy cookie in local stream.
**	Cookie should look like: "\nCookie: stunnel/kssl=4C7DD01FA175330D\n"
**	The search is not confined to HTTP headers;
**	consider this a feature to avoid leaking the random nonce ;-)
**	Return 1 if good cookie found, 0 otherwise.
*/
int 	eat_cookie(Channel *local, char *header, char *key)
{
	char	*found, *nonce, *next;
	char	*eob = &local->buff[local->ptr] - 1;
	char	backup;
	int 	i = 0, taillen;
	int 	rc = 0;

	backup = local->buff[local->ptr];		local->buff[local->ptr] = '\0';
	found = strstr(local->buff, header);	local->buff[local->ptr] = backup;
	if (!found)  return 0;

	found += strlen(header);				local->buff[local->ptr] = '\0';
	found = strstr(found, key);				local->buff[local->ptr] = backup;
	if (!found)  return 0;

	nonce = found + strlen(key);
	rc = (! options.cookie->nonce_len)?
		 0: ! strncmp(nonce, options.cookie->nonce, options.cookie->nonce_len);

	/*
	**	What happens if cookie line fragmented across BUFFSIZE buffer?
	**	Expect header to be truncated, possibly leaking nonce.
	*/
	for (next=nonce; next<eob && *next!='\n' && !isspace((int)*next); next++) ;
	while (next < eob  &&  *next != '\n'  && isspace((int)*next))  next++;
	taillen = local->ptr - (next - local->buff);
	(void) memcpy(found, next, taillen);
	local->ptr -= next - found;

	/*	may trim trailing ", "
	*/
	if (found > local->buff  &&  *found == '\n')  {
		eob = found - 1;
		for (i=0;  --eob > local->buff; i++)  {
			if (!isspace((int) *eob)  &&  *eob != ',')  break;
			}
		if (i)  {
			(void) memcpy(found-1-i, found, taillen);
			local->ptr -= i;
			}
		}

	return rc;
}


/*	Error response for bad/missing auth cookie
*/
const char	*badcookie_response =
		"HTTP/1.0 403 Access to proxy denied\n"
		"Proxy-agent: "  PROXY_VERSION  "\n"
		"Content-type: text/html\n"
		"Content-length: CONTLEN\n"
		"\n"
		"<HTML><Head><Title> Access to proxy denied</Title></Head>\n\n"
		"<Body><H1> 403 Access to proxy denied</H1>\n<P>\n"
		"Access to the klpstunnel proxy was rejected because \n"
		"your web browser did not provide the required cookie. \n"
		"Cookies are provided in the <Code> Cookie:</Code> or \n"
		"<Code> Accept-Language:</Code> HTTP headers. \n"
		"<P>\nPossible remedies:\n<OL>\n"
		"<LI>  You must run the klpstunnel proxy <Strong>before</Strong>\n"
			   "you start up your web browser\n"
		"<!--"
		"<LI>  You must enable cookies in your web browser\n"
		"--> "
		"</OL>\n"
		"</Body></HTML>\n" ;

/*	Search for cookie in local channel, set have_cookie if good one found.
**	Return 1:OK if have_cookie is set, or otherwise OK.
**	Otherwise return -1:Fail and
**		o	Write error message to local channel
**		o	Exit proxy if failed connection attempt limit exceeded.
*/
int 	got_cookie(Channel *local, Channel *remote, int km_mode)
{
	Channel		badcookie_channel;
	const int	badcookie_len = strlen(badcookie_response);
	const int	com1 = strstr(badcookie_response, "<!--") - badcookie_response;
	const int	com2 = strstr(badcookie_response, "--> ") - badcookie_response;
	char		*cp;

	/*	Always destroy cookie before sending data to remote side.
	*/
	if (eat_cookie(local, "\nCookie:", COOKIE_TAG "=")  &&
		(options.cookie->ns_auths & AUTH_COOK))  have_cookie = 1;
	if (eat_cookie(local, "\nAccept-Language:", COOKIE_TAG "-")  &&
		(options.cookie->ns_auths & AUTH_LANG))  have_cookie = 1;
	/*	Might should loop this in case multiple cookies in buffer;
	*/

	/*	Don't freak out until there's something to write to remote side.
	**	This lets the CONNECT msg and response slide by without a cookie.
	*/
	if (!remote->open  ||  !FD_ISSET(remote->fd, &wr_set))  return 1;

	if (have_cookie)  return 1;
	if (!options.cookie->nonce_len)   return 1; 	/* len=0 diasbles check	*/

	if (badcookie_len + 1 > BUFFSIZE)  {
		fprintf(stderr, "Programming botch: "
						"badcookie_response exceeds BUFFSIZE (%d)\n", BUFFSIZE);
		exit(1);
		}
	badcookie_channel.ptr = badcookie_len;
	strcpy(badcookie_channel.buff, badcookie_response);
	if ((cp = strstr(badcookie_channel.buff, "CONTLEN")))  {
		snprintf(cp, 7, "%-7d", strlen(cp+9));
		if (strlen(cp) < 10)  cp[strlen(cp)] = '\n';
		}
	if (options.cookie->ns_auths & AUTH_COOK)  {
		strncpy(&badcookie_channel.buff[com1], "    ", 4);
		strncpy(&badcookie_channel.buff[com2], "    ", 4);
		}
	FD_SET(local->fd, &wr_set);
	(void) wr_local (local, &badcookie_channel, km_mode);

	if (remote->ssl)  {
		ssl3_send_alert(remote->ssl,SSL3_AL_FATAL,SSL_AD_HANDSHAKE_FAILURE);
		}

	if (++options.cookie->tries > options.cookie->max_tries)  {
		fprintf(stderr, "proxy exiting after %d failed connection attempts\n",
				options.cookie->tries);
		  /*  printf("VRS DB: should restore prefs here?\n");	*/
		exit(1);
		}

	return -1;
}


/*	Read local channel (normally socket; ssl in kssl MitM mode)
**
**	Returns:  1:OK, 0,-1: error
*/
int 	rd_local(Channel *local, Channel *remote, int *km_mode)
{
	int 	num, retval;
	char	*hp = NULL;
	int 	in_connect = 0;

	if (!local->open  ||  !FD_ISSET(local->fd, &rd_set))  return 1;

	if (*km_mode == KM_KSSL_MITM)  {
		if ((retval = rd_ssl_ch(local, &num)))  return retval;
		}
	else if (	*km_mode == KM_LOCAL_HTTPS	||  *km_mode == KM_NORMAL  ||
				*km_mode == KM_REMOTE_HTTPS	||  *km_mode == KM_UNDEF 	)  {
		num = readsocket(local->fd, local->buff+local->ptr,BUFFSIZE-local->ptr);
		if (num < 0  &&  errno == ECONNRESET)  {
			log(LOG_NOTICE, "IPC reset (child died)");
			return 0;	/* close connection */
			}
		if (num < 0  &&  errno != EIO)  {
			sockerror("read");
			return -1;
			}
		else if (num > 0)  {
			if (*km_mode == KM_UNDEF)  {
				*km_mode = check_CONNECT(local->buff + local->ptr, &hp);
				if (*km_mode == KM_LOCAL_HTTPS  ||  *km_mode == KM_KSSL_MITM) {
					in_connect = 1;
					}
				retval = connect_ch(local, remote, *km_mode, hp);
				if (retval)   return retval;
				}

			switch (*km_mode)  {
				case KM_LOCAL_HTTPS:
				case KM_KSSL_MITM:
						if (in_connect)  break;		/*  else fall through	*/
				case KM_REMOTE_HTTPS:
				case KM_NORMAL:
						local->ptr += num;
						break;
				default:
						sockerror("check_CONNECT");
						return -1;
				}
			}
		else  { 	/* close */
			log(LOG_DEBUG, "Socket closed on read");
			local->open = 0;
			}
		}
	else  {
		sockerror("rd_local");
		return -1;
		}

	return 1;
}


/*	Read remote channel (normally ssl; socket in https proxy modes)
**
**	Returns:  1:OK; 0,-1: error
*/
int 	rd_remote(Channel *local, Channel *remote, int *km_mode)
{
	int 	num, retval;

	if (!remote->open  ||  !FD_ISSET(remote->fd, &rd_set))  return 1;

	/*	20000428 VRS Network error dialog appears in cases where rd_local
	**	fails (not ready), and we fall into rd_remote with KM_NORMAL mode.
	**	Let's try disabling this and see what happens
	*/
	if (*km_mode == KM_UNDEF)  return 1;	/* Net Error Dialog averted? */
	if (*km_mode == KM_UNDEF)  {
		*km_mode = KM_NORMAL;
		retval = connect_ch(local, remote, *km_mode, "");
		if (retval)   return retval;
		}

	if (*km_mode == KM_NORMAL  ||  *km_mode == KM_KSSL_MITM)  {
		if ((retval = rd_ssl_ch(remote, &num)))  return retval;
		}
	else if (*km_mode == KM_LOCAL_HTTPS  ||  *km_mode == KM_REMOTE_HTTPS)  {
		num = readsocket(remote->fd,
						 remote->buff + remote->ptr, BUFFSIZE - remote->ptr);
		if (num < 0  &&  errno == ECONNRESET)  {
			log(LOG_NOTICE, "IPC reset (child died)");
			return 0;	/* close connection */
			}
		if (num < 0  &&  errno != EIO)  {
			sockerror("read");
			return -1;
			}
		else if (num > 0)  {
			remote->ptr += num;
			}
		else  { 	/* close */
			log(LOG_DEBUG, "Socket closed on read");
			remote->open = 0;
			}
		}
	else  {
		sockerror("rd_remote");
		return -1;
		}

	return 1;
}


/*	Write local channel (normally socket; ssl in kssl MitM mode)
**
**	Returns:  1:OK; 0,-1: error
*/
int 	wr_local(Channel *local, Channel *remote, int km_mode)
{
	int 	num;

	if (!local->open  ||  !FD_ISSET(local->fd, &wr_set))  return 1;

	if (km_mode == KM_KSSL_MITM)  {
		wr_ssl_ch(remote, local, &num);
		}
	else if (km_mode == KM_LOCAL_HTTPS  ||  km_mode == KM_REMOTE_HTTPS  ||
			 km_mode == KM_NORMAL)  {
		num = writesocket(local->fd, remote->buff, remote->ptr);
		if (num < 0)  {
			sockerror("write");
			return -1;
			}
		if (num)  {
			memcpy(remote->buff, remote->buff+num, remote->ptr-num);
			remote->ptr  -= num;
			local->bytes += num;
			}
		else  { 	/* close */
			log(LOG_DEBUG, "Socket closed on write");
			local->open = 0;
			}
		}
	else  {
		sockerror("wr_local");
		return -1;
		}

	return 1;
}


/*	Write remote channel (normally ssl; socket in https proxy modes)
**
**	Returns:  1:OK; 0,-1: error
*/
int 	wr_remote(Channel *local, Channel *remote, int km_mode)
{
	int 	num, retval;

	if (!remote->open  ||  !FD_ISSET(remote->fd, &wr_set))  return 1;

	if (km_mode == KM_NORMAL  ||  km_mode == KM_KSSL_MITM)  {
		if ((retval = wr_ssl_ch(local, remote, &num)))  return retval;
		}
	else if (km_mode == KM_LOCAL_HTTPS  ||  km_mode == KM_REMOTE_HTTPS)  {
		num = writesocket(remote->fd, local->buff, local->ptr);
		if (num < 0)  {
			sockerror("write");
			return -1;
			}
		if (num)  {
			memcpy(local->buff, local->buff+num, local->ptr-num);
			local->ptr	  -= num;
			remote->bytes += num;
			}
		else  { 	/* close */
			log(LOG_DEBUG, "Socket closed on write");
			local->open = 0;
			}
		}
	else  {
		sockerror("wr_remote");
		return -1;
		}

	return 1;
}


/*	Check local socket data for https "CONNECT host:port HTTP/1.0" message.
**	Return mode in:
**		KM_NORMAL, KM_KSSL_MITM, KM_LOCAL_HTTPS, KM_REMOTE_HTTPS, KM_FAULT,
**	and set *hp to point to static local buffer with host:port.
*/
int 	check_CONNECT(const char *str, char **hp)
{
	static char	hostport[MAX_HP+1];
	int 		hstart, hplen;
	char 		*cp;

	hostport[0] = '\0';
	*hp = hostport;

	if (!(options.option & OPT_KHTTPS))  return KM_NORMAL;

	if (strncmp(str, "CONNECT", 7) == 0)  {
		hstart = 7 + strspn(&str[7], " \t");
		hplen = strcspn(&str[hstart], " \t");
		if (hplen >= MAX_HP)  return KM_FAULT;
		strncpy(hostport, &str[hstart], hplen);
		hostport[hplen] = '\0';
		if (strchr(*hp, ':') == NULL)  return KM_FAULT;
		}
	else  {
		return KM_NORMAL;
		}

	if (options.kssl_list  &&  (cp = strstr(options.kssl_list, *hp))  && 
		(cp[hplen] == '\0'  ||  cp[hplen] == ','))  {
		return KM_KSSL_MITM;
		}
	else if (options.https_proxy)  {
		if (strlen(options.https_proxy) >= MAX_HP)  return KM_FAULT;
		strcpy(*hp, options.https_proxy);
		return KM_REMOTE_HTTPS;
		}
	else  {
		return KM_LOCAL_HTTPS;
		}
}


/*	Fake a response to local side's "CONNECT" message
**	by pushing remote side's "200 OK" message into the remote buffer.
**	The caller should prevent delivery of local side's "CONNECT" message
**	by purposefully failing to update local->ptr + num.
*/
int 	fake_CONNECT(Channel *local)
{
	if (writesocket(local->fd, connect_response,
					strlen(connect_response)) < 0)  {
		sockerror("fake_connect");
		return 0;
		}
	return 1;
}


/*	Make new socket connection to host:port from CONNECT message.
**	Reset remote->fd from ssl fd to new fd.
**	Return 0:Success, -1:Problem
*/
int 	create_remote_sock(Channel *remote, char *hostport)
{
	struct sockaddr_in	addr;
	int 				sock;
	u_long				*host = NULL;
	u_short				port;
#if defined FIONBIO && defined USE_NBIO
	long				l;
#endif

	if (!remote  ||  !hostport)  {
		sockerror("create_remote_sock args");
		return -1;
		}

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
		sockerror("remote hp socket");
		return -1;
		}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	name2nums(hostport, &host, &port);
	addr.sin_port = port;
	addr.sin_addr.s_addr = *host;

	if(!connect(sock, (struct sockaddr *) &addr, sizeof(addr)))  {
#if defined FIONBIO && defined USE_NBIO
		l = 1; /* ON */
		if(ioctlsocket(sock, FIONBIO, &l)<0)
			sockerror("ioctlsocket (remote sock)"); /* non-critical */
		log(LOG_DEBUG, "Remote socket set to non-blocking mode");
#endif
		remote->fd  = sock;
		return 0;		/* success */
		}
	else  {
		sockerror("remote hp connect");
		return -1;
		}
}


#ifndef NO_RSA
static RSA *tmp_rsa_cb(SSL *s, int export, int keylength)
{ /* temporary RSA key callback */
    log(LOG_DEBUG, "Returned temporary RSA callback");
    return rsa_tmp;
}
#endif /* NO_RSA */

static int verify_callback(int state, X509_STORE_CTX *ctx)
{ /* our verify callback function */
    char txt[256];
    X509_OBJECT ret;

    X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),
        txt, sizeof(txt));
    if(!state) {
        /* Remote site specified a certificate, but it's not correct */
        log(LOG_WARNING, "VERIFY ERROR: depth=%d error=%s: %s",
            ctx->error_depth,
            X509_verify_cert_error_string (ctx->error), txt);
        return 0; /* Reject connection */
    }
    if(options.verify_use_only_my && ctx->error_depth==0 &&
            X509_STORE_get_by_subject(ctx, X509_LU_X509,
                X509_get_subject_name(ctx->current_cert), &ret)!=1) {
        log (LOG_WARNING, "VERIFY ERROR ONLY MY: no cert for: %s", txt);
        return 0; /* Reject connection */
    }
    log(LOG_NOTICE, "VERIFY OK: depth=%d: %s", ctx->error_depth, txt);
    return 1; /* Accept connection */
}

static void info_callback(SSL *s, int where, int ret)
{
    log(LOG_DEBUG, "%s", SSL_state_string_long(s));
    if(where==SSL_CB_HANDSHAKE_DONE)
        print_stats(s->ctx);
}

static void print_stats(SSL_CTX *ctx) /* print statistics */
{
    log(LOG_DEBUG, "%4ld items in the session cache",
        SSL_CTX_sess_number(ctx));
    log(LOG_DEBUG, "%4d client connects (SSL_connect())",
        SSL_CTX_sess_connect(ctx));
    log(LOG_DEBUG, "%4d client connects that finished",
        SSL_CTX_sess_connect_good(ctx));
#if SSLEAY_VERSION_NUMBER >= 0x0922
    log(LOG_DEBUG, "%4d client renegotiatations requested",
        SSL_CTX_sess_connect_renegotiate(ctx));
#endif
    log(LOG_DEBUG, "%4d server connects (SSL_accept())",
        SSL_CTX_sess_accept(ctx));
    log(LOG_DEBUG, "%4d server connects that finished",
        SSL_CTX_sess_accept_good(ctx));
#if SSLEAY_VERSION_NUMBER >= 0x0922
    log(LOG_DEBUG, "%4d server renegotiatiations requested",
        SSL_CTX_sess_accept_renegotiate(ctx));
#endif
    log(LOG_DEBUG, "%4d session cache hits", SSL_CTX_sess_hits(ctx));
    log(LOG_DEBUG, "%4d session cache misses", SSL_CTX_sess_misses(ctx));
    log(LOG_DEBUG, "%4d session cache timeouts", SSL_CTX_sess_timeouts(ctx));
}

static void print_cipher(SSL *ssl) /* print negotiated cipher */
{
#if SSLEAY_VERSION_NUMBER > 0x0800
    SSL_CIPHER *c;
    char *ver;
    int bits;
#endif

#if SSLEAY_VERSION_NUMBER <= 0x0800
    log(LOG_INFO, "%s opened with SSLv%d, cipher %s",
        options.servname, ssl->session->ssl_version, SSL_get_cipher(ssl));
#else
    switch(ssl->session->ssl_version) {
    case SSL2_VERSION:
        ver="SSLv2"; break;
    case SSL3_VERSION:
        ver="SSLv3"; break;
    case TLS1_VERSION:
        ver="TLSv1"; break;
    default:
        ver="UNKNOWN";
    }
    c=SSL_get_current_cipher(ssl);
    SSL_CIPHER_get_bits(c, &bits);
    log(LOG_INFO, "%s opened with %s, cipher %s (%u bits)",
        options.servname, ver, SSL_CIPHER_get_name(c), bits);
#endif
}

static void sslerror(char *txt) /* SSL Error handler */
{
    char string[120];

    ERR_error_string(ERR_get_error(), string);
    log(LOG_ERR, "%s: %s", txt, string);
}

/* End of ssl.c */

